#ifndef AGGREGATION_H
#define AGGREGATION_H

#include "intlist.h"
#include "ugraph.h"
#include "utree.h"
#include "rbtree.h"
#include "bandwidth.h"

#define AGGREGATION_INFINITY 1000000.0
#define AGGREGATION_BITMAP long long
#define AGGREGATION_EDGEROUTERS_FOR_RANDOM 60
//#define AGGREGATION_NOTREE //AGGREGATION_NOTREE: no trees are stored in the aggregation (to make it faster)

struct aggregation {
	struct ugraph * graph;
	struct bandwidth * bandwidth;
	int numberEdgeRouters;

	// groups
	int maxNumberOfGroups;
	int numberOfGroups;
	int numberOfAcceptedGroups;
	AGGREGATION_BITMAP * groups;

	// trees
	int maxNumberOfTrees;
  int numberOfTrees;
  int change;
#ifdef AGGREGATION_NOTREE
	int * treecost;
	int * treenumberofspannednodes;
	int * treenumberofedges;
#else
	struct utree ** trees;
#endif
	AGGREGATION_BITMAP * treebitmap;
	// match each group to a tree
	int * groupToTree;
	int * numberOfAggregatedGroups;
  int * groupRequirements;
  int * nativeTreeCost;
  // number of calls to the "comparison" function
	long long numberOfComparisons;
	int maxNumberOfComparisons;
  struct intlist ** costset;
};


/*main*/
void aggregationMain(int argc,char ** argv);

/*************************************************************************************************/
/******************* Creation and Destruction ****************************************************/
/*************************************************************************************************/

/* creates a new aggregation structure for at most "maxNumberOfGroups" groups and at most "maxNumberOfTrees" trees */
struct aggregation * aggregationCreate(struct ugraph * graph,int maxNumberOfGroups,int maxNumberOfTrees,int numberEdgeRouters,struct bandwidth * bandwidth);

/*free the aggregation structure*/
void aggregationFree(struct aggregation * a);


/*************************************************************************************************/
/*******************      Display functions   ****************************************************/
/*************************************************************************************************/


/* displays "a": the total time to make all the aggregations, the percentage of aggregations, the number of forwarding entries in the routing table, the size of the forawarding table, and finally the average bandwidth used per group*/
void aggregationDisplay(struct aggregation * a);

/* display the size of "costset" sets: the number of trees of the same cost  */
void aggregationGetTreesPerCost(struct aggregation * a);

/* returns the standard deviation of used edges "a" */
float aggregationGetStandardDeviationEdges(struct aggregation * a);

/*returns the total number of members in the network*/
int aggregationGetTotalNumberOfMembers(struct aggregation * a);

/*returns the bandwidth used per group if no aggregation (i.e. the minimum bandwidth needed for all the groups as in traditional IP multicast)*/
int aggregationGetNativeBandwidth(struct aggregation * a);
float aggregationGetGroupRequirements(struct aggregation * a);
/* returns the total bandwidth used for all the groups*/
int aggregationGetTotalBandwidth(struct aggregation * a);

/* returns the total number of forwarding entries : how many entries are created in the routing table for all the routing tables*/
int aggregationGetNumberOfForwardingEntries(struct aggregation * a);

/* returns the total size of forwarding table, a forwarding entry has a size corresponding to the cost of a tree.  */
int aggregationGetSizeOfForwardingTable(struct aggregation * a);

/* displays link usage */
void aggregationLinkUsage(struct aggregation * a);

/*************************************************************************************************/
/*******************   STA functions : Aggregation ***********************************************/
/*************************************************************************************************/


/* adds a new tree covering "group" to "a" using minimum spanning tree or shortest path tree algorithm*/
void aggregationAddNewTree(struct aggregation * a,AGGREGATION_BITMAP group);

/* aggregate a group (the current group is "a->numberOfGroups") to a tree "tree" in "a" */
void aggregationAggregateToTree(struct aggregation * a,int tree);

/* adds "group" to "a" using STA and then decide whether or not to aggregate it to a tree*/
int aggregationAddGroupSTA1(struct aggregation * a,AGGREGATION_BITMAP group,int groupRequirement); // STA envoy  ICC
int aggregationAddGroupSTA2(struct aggregation * a,AGGREGATION_BITMAP group,int groupRequirement); // STA avec "bandwidthCheck"
int aggregationAddGroupSTA3(struct aggregation * a,AGGREGATION_BITMAP group,int groupRequirement); // STA avec "bandwidthCheck" et "tree_g" dans le "graphCopy"

/*adds "group" to "a" using STACongestion and then decide whether or not to aggregate it to a tree*/
int aggregationAddGroupStaCongestion(struct aggregation * a,AGGREGATION_BITMAP group,int groupRequirement); 

/*Returns a Candidate among all the candidates in "candidates"*/
int aggregationCandidate(int * candidates);

/* adds "group" to "a" without aggregation as in traditional IP multicast*/
int aggregationAddGroupNoAggregation(struct aggregation * a,AGGREGATION_BITMAP group,int groupRequirement);

void aggregationDisplayStress(struct aggregation * a);
void aggregationAddTree(struct aggregation * a,AGGREGATION_BITMAP group,struct utree * tree_g);
float delta_AQOSM(struct aggregation * a,struct intlist * group,int groupRequirement,int tree_0);
int aggregationAddGroupAQOSM(struct aggregation * a,AGGREGATION_BITMAP group,int groupRequirement);

/*************************************************************************************************/
/*******************   STA functions : Bitmaps     ***********************************************/
/*************************************************************************************************/


/* transforms the bitmap "bitmap" into a list of members */
struct intlist * aggregationBitmapToIntlist(AGGREGATION_BITMAP group);

/* transforms the list of members "group" into a bitmap */
AGGREGATION_BITMAP aggregationIntlistToBitmap(struct intlist * group);

/* returns the bitmap corresponding to the nodes covered by "tree" */
AGGREGATION_BITMAP aggregationTreeBitmap(struct utree * tree,int numberEdgeRouters);

/*******************************/
/***** Node Weighted Model *****/
/*******************************/

struct nodeWeightedModel {
	int * nodeWeight;
	int nodeNumber;
};

struct nodeWeightedModel * nodeWeightedModelCreate(int nodeNumber);
void nodeWeightedModelFree(struct nodeWeightedModel * model);
struct intlist * nodeWeightedModelNewGroup(struct nodeWeightedModel * model,int groupSize);

#endif
