#include "alist.h"
#include "intlist.h"
#include "qerror.h"
#include <stdio.h>

/* creates an empty alist */
struct alist * alistCreate(void (* alistFreeElement)(void *),void * (alistCopyElement)(void *),void (* alistDisplayElement)(void *)) {
	struct alist * l;
	QERROR_MALLOC(l,1,struct alist);
	l->keys = intlistCreate();
	l->values = listCreate(alistFreeElement,alistCopyElement,alistDisplayElement);
	return l;
}

/* frees "l" */
void alistFree(struct alist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	intlistFree(l->keys);
	listFree(l->values);
	free(l);
}

/* copies "l" */
struct alist * alistCopy(struct alist * l) {
	struct alist * copy;
	if (l==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(copy,1,struct alist);
	copy->keys = intlistCopy(l->keys);
	copy->values = listCopy(l->values);
	return copy;
}

/* returns the length of "l" */
int alistLength(struct alist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return intlistLength(l->keys);
}

/* returns the size of "l" */
int alistSize(struct alist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return intlistSize(l->keys);
}

/* checks if "l" is empty */
int alistIsEmpty(struct alist * l) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return intlistIsEmpty(l->keys);
}

/* returns the value in "l" associated to "key" */
void * alistGet(struct alist * l,int key) {
	struct intlist_element * key_element;
	struct list_element * value_element;
	int found;
	if (l==NULL) {
		QERROR("parameters");
	}
	found = 0;
	key_element = l->keys->head;
	value_element = l->values->head;
	while ((!found) && (key_element!=NULL)) {
		if (key_element->value==key) {
			found = 1;
		}
		else {
			key_element = key_element->next;
			value_element = value_element->next;
		}
	}
	if (found) {
		return value_element->value;
	}
	else {
		return NULL;
	}
}

/* checks if "l" contains "key" */
int alistContain(struct alist * l,int key) {
	if (l==NULL) {
		QERROR("parameters");
	}
	return intlistContain(l->keys,key);
}

/* returns the number of entries of "l" having key "key" */
int alistNumber(struct alist * l,int key) {
	struct intlist_element * element;
	int number;
	if (l==NULL) {
		QERROR("parameters");
	}
	number = 0;
	element = l->keys->head;
	while (element!=NULL) {
		if (element->value==key) {
			number = number+1;
		}
		element = element->next;
	}
	return number;
}

/* removes all entries of "l" having key "key" */
void alistRemoveAll(struct alist * l,int key) {
	struct intlist_element * key_element,* tmp_key;
	struct list_element * value_element,* tmp_value;
	int found;
	if (l==NULL) {
		QERROR("parameters");
	}
	found = 0;
	key_element = l->keys->head;
	value_element = l->values->head;
	while (key_element!=NULL) {
		tmp_key = key_element->next;
		tmp_value = value_element->next;
		if (key_element->value==key) {
			found = 1;
			intlistDelete(l->keys,key_element);
			listDelete(l->values,value_element);
		}
		key_element = tmp_key;
		value_element = tmp_value;
	}
	if (!found) {
		QERROR("not in alist");
	}
}

/* removes the entry ("key","value") in "l", using function "compare" */
void alistRemoveParticular(struct alist * l,int key,void * value,int (* compare)(void *,void *)) {
	struct intlist_element * key_element;
	struct list_element * value_element;
	int found;
	if (l==NULL) {
		QERROR("parameters");
	}
	found = 0;
	key_element = l->keys->head;
	value_element = l->values->head;
	while ((!found) && (key_element!=NULL)) {
		if ((key_element->value==key) && (compare(value_element->value,value)==0)) {
			found = 1;
		}
		else {
			key_element = key_element->next;
			value_element = value_element->next;
		}
	}
	if (found) {
		intlistDelete(l->keys,key_element);
		listDelete(l->values,value_element);
	}
	else {
		QERROR("not in alist");
	}
}

/* removes the entry ("key","value") in "l" */
void alistRemoveParticularStandard(struct alist * l,int key,void * value) {
	alistRemoveParticular(l,key,value,listFindCompareEqual);
}

/* adds ("key","value") to "l" (allowing multiple keys) */
void alistAdd(struct alist * l,int key,void * value) {
	if (l==NULL) {
		QERROR("parameters");
	}
	intlistAddToFront(l->keys,key);
	listAddToFront(l->values,value);
}

/* sorts "l" according to the keys */
void alistSort(struct alist * l) {
	struct intlist * sortedKeyList;
	struct intlist_element * key_element;
	struct intlist_element * key_minimum_element;
	struct list * sortedValuesList;
	struct list_element * values_element;
	struct list_element * values_minimum_element;
	sortedKeyList = intlistCreate();
	sortedValuesList = listCreate(l->values->listFreeElement,l->values->listCopyElement,l->values->listDisplayElement);
	while (!intlistIsEmpty(l->keys)) {
		key_minimum_element = l->keys->head;
		values_minimum_element = l->values->head;
		key_element = l->keys->head->next;
		values_element = l->values->head->next;
		while (key_element!=NULL) {
			if (key_element->value<key_minimum_element->value) {
				key_minimum_element = key_element;
				values_minimum_element = values_element;
			}
			key_element = key_element->next;
			values_element = values_element->next;
		}
		intlistAddToBack(sortedKeyList,key_minimum_element->value);
		listAddToBack(sortedValuesList,values_minimum_element->value);
		intlistDelete(l->keys,key_minimum_element);
		listDelete(l->values,values_minimum_element);
	}
	l->keys->head = sortedKeyList->head;
	l->values->head = sortedValuesList->head;
	l->keys->tail = sortedKeyList->tail;
	l->values->tail = sortedValuesList->tail;
	l->keys->size = sortedKeyList->size;
	l->values->size = sortedValuesList->size;
	free(sortedKeyList);
	free(sortedValuesList);
}

/* displays "l" */
void alistDisplay(struct alist * l) {
	struct intlist_element * key_element;
	struct list_element * value_element;
	if (l==NULL) {
		QERROR("parameters");
	}
	key_element = l->keys->head;
	value_element = l->values->head;
	printf("( ");
	while (key_element!=NULL) {
		printf("(%d,",key_element->value);
		l->values->listDisplayElement(value_element->value);
		printf(") ");
		key_element = key_element->next;
		value_element = value_element->next;
	}
	printf(")");
}
