#include "utree.h"
#include "qerror.h"
#include "intstack.h"
#include <stdio.h>
/* creates an empty utree, having "source" as its roots, from "support" */
struct utree * utreeCreate(struct ugraph * support,int source) {
	struct utree * t;
	if ((support==NULL) || (source<0) || (source>=support->nodeNumber)) {
		QERROR("parameters");
	}
	QERROR_MALLOC(t,1,struct utree);
	t->source = source;
	t->tree = ugraphCreate(support->nodeNumber,support->ugraphFreeElement,support->ugraphCopyElement,support->ugraphDisplayElement);
	return t;
}

/* creates from ugraph "support" a shortest path tree of "members" and having "source" as its root */
struct utree * utreeCreateShortestPathTree(struct ugraph * support,int source,struct intlist * members) {
	struct utree * t;
	int * predecessors,* distances;
	struct intlist_element * nextMember;
	int connected = 1;
	if ((support==NULL) || (source<0) || (source>=support->nodeNumber) || (members==NULL)) {
		QERROR("parameters");
	}
	t = utreeCreate(support,source);
	QERROR_MALLOC(predecessors,support->nodeNumber,int);
	QERROR_MALLOC(distances,support->nodeNumber,int);
	ugraphDijkstra(support,source,predecessors,distances);
	nextMember = members->head;
	while (nextMember!=NULL) {
		//printf("nextMember->value=%d predecessor[%d]=%d\n",nextMember->value,nextMember->value,predecessors[nextMember->value]);
		if (predecessors[nextMember->value]==-1) {
			connected = 0;
		}
		else {
			utreeAddNode(t,support,nextMember->value,predecessors);	
		}
		nextMember = nextMember->next;
		
	}
	free(predecessors);
	free(distances);
	if (!connected) {
		utreeFree(t);
		return NULL;
	}
	else {
		return t;
	}
}

struct utree * utreeCreateMinimumSpanningTree(struct ugraph * support, int source, struct intlist * members){
  struct utree * t;
  int * predecessors,* distances;
  int end = 0;
  int v;
  int connected = 1;
  //int i,j;
  struct intlist_element * element;
  if ((support==NULL) || (source<0) || (source>=support->nodeNumber) || (members==NULL)) {
    QERROR("parameters");
  }

  t = utreeCreate(support,source);
  QERROR_MALLOC(predecessors,support->nodeNumber,int);
  QERROR_MALLOC(distances,support->nodeNumber,int);
  
  while(end<(members->size - 1)){ 
    
    utreeDijkstra(t,support,members,predecessors,distances);
    v = utreeGetClosestMember(t,members,distances);
    //printf("end =%d et v=%d\n",end,v);
    if(v==-1) {
      connected = 0;
      end = members->size;
    }
    else {
      element = intlistSearch(members,v);
      utreeAddNode(t,support,element->value,predecessors);
      end++;
    }    
  }
  free(predecessors);
  free(distances);
 
  if (connected) {
 
    return t;
  }
  else {
 
    utreeFree(t);
    return NULL;
  }
}

struct utree * utreeCreateSteinerTree1Member(struct ugraph * support,int source,struct intlist * members) {
	struct utree * steiner;
	int * distances;
	int * predecessors;
	if ((support==NULL) || (source<0) || (members==NULL) || (intlistSize(members)!=1)) {
		QERROR("parameters");
	}
	QERROR_MALLOC(distances,support->nodeNumber,int);
	QERROR_MALLOC(predecessors,support->nodeNumber,int);
	ugraphDijkstra(support,source,predecessors,distances);
	steiner = utreeCreate(support,source);
	utreeAddNode(steiner,support,members->head->value,predecessors);
	free(predecessors);
	free(distances);
	return steiner;
}

struct utree * utreeCreateSteinerTree2Members(struct ugraph * support,int source,struct intlist * members) {
	struct utree * steiner;
	int min_b,min;
	int b,d1,d2;
	int * distances;
	int * predecessors;
	if ((support==NULL) || (source<0) || (members==NULL) || (intlistSize(members)!=2)) {
		QERROR("parameters");
	}
	d1 = members->head->value;
	d2 = members->head->next->value;
	min_b = -1;
	min = UGRAPH_INFINITY;
	QERROR_MALLOC(distances,support->nodeNumber,int);
	for (b=0;b<support->nodeNumber;b++) {
		ugraphDijkstra(support,b,NULL,distances);
		if (distances[source]+distances[d1]+distances[d2]<min) {
			min_b = b;
			min = distances[source]+distances[d1]+distances[d2];
		}
	}
	QERROR_MALLOC(predecessors,support->nodeNumber,int);
	ugraphDijkstra(support,source,predecessors,distances);
	steiner = utreeCreate(support,source);
	utreeAddNode(steiner,support,min_b,predecessors);
	ugraphDijkstra(support,min_b,predecessors,distances);
	utreeAddNode(steiner,support,d1,predecessors);
	utreeAddNode(steiner,support,d2,predecessors);
	free(predecessors);
	free(distances);
	return steiner;
}

struct utree * utreeCreateSteinerTree3Members(struct ugraph * support,int source,struct intlist * members) {
	struct utree * steiner;
	int min_b1,min_b2,min_d1,min_d2,min_d3,min;
	int b1,b2,d1,d2,d3;
	int * distances1,* distances2;
	int * predecessors;
	if ((support==NULL) || (source<0) || (members==NULL) || (intlistSize(members)!=3)) {
		QERROR("parameters");
	}
	d1 = members->head->value;
	d2 = members->head->next->value;
	d3 = members->head->next->next->value;
	min_b1 = -1;
	min_b2 = -1;
	min_d1 = -1;
	min_d2 = -1;
	min_d3 = -1;
	min = UGRAPH_INFINITY;
	QERROR_MALLOC(distances1,support->nodeNumber,int);
	QERROR_MALLOC(distances2,support->nodeNumber,int);
	for (b1=0;b1<support->nodeNumber;b1++) {
		ugraphDijkstra(support,b1,NULL,distances1);
		for (b2=0;b2<support->nodeNumber;b2++) {
			ugraphDijkstra(support,b2,NULL,distances2);
			if (distances1[source]+distances1[d1]+distances1[b2]+distances2[d2]+distances2[d3]<min) {
				min_b1 = b1;
				min_b2 = b2;
				min_d1 = d1;
				min_d2 = d2;
				min_d3 = d3;
				min = distances1[source]+distances1[d1]+distances1[b2]+distances2[d2]+distances2[d3];
			}
			if (distances1[source]+distances1[d2]+distances1[b2]+distances2[d1]+distances2[d3]<min) {
				min_b1 = b1;
				min_b2 = b2;
				min_d1 = d2;
				min_d2 = d1;
				min_d3 = d3;
				min = distances1[source]+distances1[d2]+distances1[b2]+distances2[d1]+distances2[d3];
			}
			if (distances1[source]+distances1[d3]+distances1[b2]+distances2[d2]+distances2[d1]<min) {
				min_b1 = b1;
				min_b2 = b2;
				min_d1 = d3;
				min_d2 = d2;
				min_d3 = d3;
				min = distances1[source]+distances1[d3]+distances1[b2]+distances2[d2]+distances2[d1];
			}
		}
	}
	QERROR_MALLOC(predecessors,support->nodeNumber,int);
	steiner = utreeCreate(support,source);
	ugraphDijkstra(support,source,predecessors,distances1);
	utreeAddNode(steiner,support,min_b1,predecessors);
	ugraphDijkstra(support,min_b1,predecessors,distances1);
	utreeAddNode(steiner,support,min_d1,predecessors);
	utreeAddNode(steiner,support,min_b2,predecessors);
	ugraphDijkstra(support,min_b2,predecessors,distances1);
	utreeAddNode(steiner,support,min_d2,predecessors);
	utreeAddNode(steiner,support,min_d3,predecessors);
	free(predecessors);
	free(distances2);
	free(distances1);
	return steiner;
}

struct utree * utreeCreateSteinerTree(struct ugraph * support, int source, struct intlist * members){
  struct utree * mintree=NULL;
  int mincost = UGRAPH_INFINITY;
  int i,compteur,maxcompteur,j,compteur_save;
  int * nonmembers;
  struct intlist * steinernodes;
  struct utree * tree = NULL;
  if ((support==NULL) || (source<0) || (source>=support->nodeNumber) || (members==NULL)) {
    QERROR("parameters");
  }
	if (intlistSize(members)==1) {
		return utreeCreateSteinerTree1Member(support,source,members);
	}
	if (intlistSize(members)==2) {
		return utreeCreateSteinerTree2Members(support,source,members);
	}
	if (intlistSize(members)==3) {
		return utreeCreateSteinerTree3Members(support,source,members);
	}
  QERROR_MALLOC(nonmembers,(support->nodeNumber)-(members->size)-1,int);
  j = 0;
  for (i=0;i<support->nodeNumber;i++) {
    if ((i!=source) && (!intlistContain(members,i))) {
      nonmembers[j] = i;
      j = j+1;
    }
  }
  maxcompteur = (1<<(support->nodeNumber-members->size-1))-1;

  for (compteur=0;compteur<=maxcompteur;compteur++) {
    steinernodes = intlistCopy(members);
    j = 0;
    compteur_save = compteur;
    while (compteur!=0) {
      if (compteur%2==1) {
        intlistPush(steinernodes,nonmembers[j]);
      }
      j  = j+1;
      compteur = compteur/2;
    }
    compteur = compteur_save;
    intlistPush(steinernodes,source);

    tree = utreeCreateMinimumSpanningTree(support,source,steinernodes);
    if (utreeTotalCost(tree)<mincost) {
      mincost = utreeTotalCost(tree);
      if (mintree!=NULL) {
	utreeFree(mintree);
      }
      mintree = tree;
    }
    else {
      utreeFree(tree);
    }
    intlistFree(steinernodes);
  }
  free(nonmembers);
  return mintree;
}


/* frees "t" */
void utreeFree(struct utree * t) {
	if (t==NULL) {
		QERROR("parameters");
	}
	ugraphFree(t->tree);
	free(t);
}

/* displays "t" */
void utreeDisplay(struct utree * t) {
	if (t==NULL) {
		QERROR("parameters");
	}
	ugraphDisplay(t->tree);
}

/* copies "t" */
struct utree * utreeCopy(struct utree * t) {
	struct utree * copy;
	if (t==NULL) {
		QERROR("parameters");
	}
	QERROR_MALLOC(copy,1,struct utree);
	copy->source = t->source;
	copy->tree = ugraphCopy(t->tree);
	return copy;
}

/* adds the edge ("node1","node2") and the edge ("node2","node1") in "t" with cost "cost" (a copy) */
void utreeAddEdge(struct utree * t,int node1,int node2,void * cost) {
	if ((t==NULL) || (node1<0) || (node1>=t->tree->nodeNumber) || (node2<0) || (node2>=t->tree->nodeNumber)) {
		QERROR("parameters");
	}
	if (((utreeCoverNode(t,node1)) && (!utreeCoverNode(t,node2))) || ((!utreeCoverNode(t,node1)) && (utreeCoverNode(t,node2)))) {
		ugraphAddEdge(t->tree,node1,node2,cost);
	}
	else {
		QERROR("not a tree");
	}
}

/* checks if "t" covers "node" */
int utreeCoverNode(struct utree * t,int node) {
  if(t==NULL){
    QERROR("tree NULL");
  }
  if(node<0){
    printf("utreeCoverNode: node=%d\n",node);
    QERROR("node");
  }
	if ((node<0) || (node>=t->tree->nodeNumber)) {
		QERROR("parameters");
	}
	if ((node==t->source) || (ugraphGetNodeDegree(t->tree,node)>=1)) {
		return 1;
	}
	else {
		return 0;
	}
}

/* checks if "t" covers all the elements of "nodes" */
int utreeCoverListOfNodes(struct utree * t,struct intlist * nodes) {
  int cover = 1;
  struct intlist_element * node;
  if ((t==NULL) || (nodes==NULL)) {
    QERROR("parameters");
  }
  node = nodes->head;
  while ((cover==1) && (node!=NULL)) {
    if (!utreeCoverNode(t,node->value)) {
      cover = 0;
    }
    node = node->next;
  }
  return cover;
}

/* checks if ("node1","node2") is in "t" */
int utreeContainEdge(struct utree * t,int node1,int node2) {
	if ((t==NULL) || (node1<0) || (node1>=t->tree->nodeNumber) || (node2<0) || (node2>=t->tree->nodeNumber)) {
		printf("t=%p node1=%d node2=%d t->tree->nodeNumber=%d\n",t,node1,node2,t->tree->nodeNumber);
		QERROR("parameters");
	}
	return ugraphIsEdge(t->tree,node1,node2);
}

void utreeDijkstra(struct utree * t,struct ugraph * support,struct intlist * members,int * predecessors,int * distances) {
	int i;
	if ((t==NULL) || (support==NULL) || (members==NULL)) {
		QERROR("parameters");
	}
	for (i=0;i<t->tree->nodeNumber;i++){

	  if (utreeCoverNode(t,i)) {
			distances[i] = 0;
			
			if (predecessors!=NULL) {
				predecessors[i] = i;
			}
		}
		else {
			distances[i] = UGRAPH_INFINITY;
			if (predecessors!=NULL) {
				predecessors[i] = -1;
			}
		}
	}
	ugraphDijkstraNoInit(support,predecessors,distances);
}

/* searchs for the closest unspanned member of "members" in "t", using the "distances" array */
int utreeGetClosestMember(struct utree * t,struct intlist * members,int * distances) {
	struct intlist_element * nextMember;
	int closestMember;
	int minimumDistance;
	if ((t==NULL) || (members==NULL) || (distances==NULL)) {
		QERROR("parameters");
	}
	minimumDistance = UGRAPH_INFINITY;
	closestMember = -1;
	nextMember = members->head;
	while (nextMember!=NULL) {
		if (!utreeCoverNode(t,nextMember->value)) {
			if (distances[nextMember->value]<minimumDistance) {
				closestMember = nextMember->value;
				minimumDistance = distances[nextMember->value];
			}
		}
		nextMember = nextMember->next;
	}
	return closestMember;
}

/* adds "node" to "t" using the "predecessors" array */
void utreeAddNode(struct utree * t,struct ugraph * support,int node,int * predecessors) {
	struct intstack * firstNode;
	struct intstack * secondNode;
	int node1,node2;
	if ((t==NULL) || (support==NULL) || (node<0) || (node>=t->tree->nodeNumber) || (predecessors==NULL)) {
		QERROR("parameters");
	}
	firstNode = intstackCreate();
	secondNode = intstackCreate();
	//printf("dans utreeaddnode %d \n\n\n\n\n",node);
	while (!utreeCoverNode(t,node)) {
	  //printf("dans utreeaddnode");
		intstackPush(firstNode,node);
		intstackPush(secondNode,predecessors[node]);
		node = predecessors[node];
	}
	while (!intstackIsEmpty(firstNode)) {
		node1 = intstackPop(firstNode);
		node2 = intstackPop(secondNode);
		utreeAddEdge(t,node1,node2,ugraphGetEdgeCost(support,node1,node2));
	}
	intstackFree(firstNode);
	intstackFree(secondNode);
}



/* returns the total cost of "t" */
int utreeTotalCost(struct utree * t) {
  struct list_element * edgevalues;
  struct intlist_element * neighbours;
  int i;
  int cost;
  if (t==NULL) {
    QERROR("parameters");
  }
  cost = 0;
  for (i=0;i<t->tree->nodeNumber;i++) {
    neighbours = t->tree->nodes[i]->keys->head;
    edgevalues = t->tree->nodes[i]->values->head;
    while (edgevalues!=NULL) {
      cost = cost+ugraphIntEdgeCost(edgevalues->value);
      edgevalues = edgevalues->next;
      neighbours = neighbours->next;
    }
  }
  return cost/2;
}

/* returns the number of nodes spanned by "t" */
int utreeNumberOfSpannedNodes(struct utree * t) {
	int i,number;
	if (t==NULL) {
		QERROR("parameters");
	}
	number = 0;
	for (i=0;i<t->tree->nodeNumber;i++) {
		if (utreeCoverNode(t,i)) {
			number = number+1;
		}
	}
	return number;
}

/* returns the number of branching nodes of "t" (the source is a branching node if it has a degree 2 or greater, other nodes are branching nodes if they have a degree 3 or greater) */
int utreeNumberOfBranchingNodes(struct utree * t) {
	int number,i;
	if (t==NULL) {
		QERROR("parameters");
	}
	number = 0;
	for (i=0;i<t->tree->nodeNumber;i++) {
		if (((i==t->source) && (ugraphGetNodeDegree(t->tree,i)>=2)) || (ugraphGetNodeDegree(t->tree,i)>=3)) {
			number = number+1;
		}
	}
	return number;
}

/* returns the number of edges of "t" */
int utreeNumberOfEdges(struct utree * t) {
	if (t==NULL) {
		QERROR("parameters");
	}
	return utreeNumberOfSpannedNodes(t)-1;
}

/*returns the number of leaves of "t" (the source of "t" can be a leaf) */
int utreeNumberOfLeaves(struct utree * t) {
  int i;
  int leaves;
  if (t==NULL) {
    QERROR("parameters");
  }
  leaves = 0;
  for (i=0;i<t->tree->nodeNumber;i++) {
    if (ugraphGetNodeDegree(t->tree,i)==1){
     leaves = leaves+1;
    }
  }
  return leaves;
}


/* returns the parent of "node" in "t" */
int utreeParent(struct utree * t,int node) {
	struct intlist * nodestack;
	struct intlist * previousstack;
	struct intlist_element * e;
	int n,previous;
	int found;
	if ((t==NULL) || (node<0) || (node>=t->tree->nodeNumber)) {
		QERROR("parameters");
	}
	nodestack = intlistCreate();
	previousstack = intlistCreate();
	e = t->tree->nodes[t->source]->keys->head;
	while (e!=NULL) {
		intlistPush(nodestack,e->value);
		intlistPush(previousstack,t->source);
		e = e->next;
	}
	previous = t->source;
	found = 0;
	while ((!found) && (!intlistIsEmpty(nodestack))) {
		n = intlistPop(nodestack);
		previous = intlistPop(previousstack);
		if (n==node) {
			found = 1;
		}
		else {
			e = t->tree->nodes[n]->keys->head;
			while (e!=NULL) {
				if (e->value!=previous) {
					intlistPush(nodestack,e->value);
					intlistPush(previousstack,n);
				}
				e = e->next;
			}
		}
	}
	intlistFree(nodestack);
	intlistFree(previousstack);
	return previous;
}


























/*
struct intlist * utreeBreadthFirstSearch(struct utree * t) {
	if (t==NULL) {
		QERROR("parameters");
	}
}
*/

/*
struct intlist * utreeSubtree(struct utree * t,int node) {
	if (t==NULL) {
		QERROR("parameters");
	}
}
*/

/*
void utreeContractInGraph(struct utree * t,struct ugraph * g,int node) {
	if (t==NULL) {
		QERROR("parameters");
	}
}
*/
